// This implementation file includes all the implementation files needed
// to get MemTailor working.
#include "mathic.h"
#include "../../src/Timer.cpp"
#include "../../src/ColumnPrinter.cpp"
#include "../../src/DivMask.cpp"
#include "../../src/Action.cpp"
#include "../../src/BoolParameter.cpp"
#include "../../src/CliParameter.cpp"
#include "../../src/CliParser.cpp"
#include "../../src/NameFactory.cpp"
#include "../../src/error.cpp"
#include "../../src/HelpAction.cpp"
#include "../../src/IntegerParameter.cpp"
#include "../../src/StringParameter.cpp"
#include "../../src/display.cpp"
